#pragma TextEncoding	= "UTF-8"
#pragma rtGlobals		= 3
#pragma IgorVersion		= 8.00
#pragma moduleName		= AnalysisTools
#pragma version			= 1.11
#pragma DefaultTab		= {3,20,4}

static Constant kVersion = 1.11
static StrConstant kVersionDate = "11/2023"

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static Constant kProjectID = 21866																// the project node on IgorExchange
static StrConstant ksShortTitle = "Super Quick Fit"												// the project short title on IgorExchange

//________________________________________________________________________________________________
//										Super Quick Fit
//	Written by Stephan Thuermer - https://www.wavemetrics.com/user/chozo
//
//	Provides a new right-click menu similar to the built-in Quick Fit. Several custom options have
//	been added for quick access and user-defined functions are supported. To add your own function
//	you need to provide:
//		- a standard or all-at-once fit function with an unique name
//		- a guess function with the name fitfuncName+CoefGen_SUFFIX
//	
//	Optionally, derived results can be printed from another function named fitfuncName+derivedVal_SUFFIX
//
//	Both the guess and derived-results functions have the SuperQuickFitStruct as input.
//	The guess function is expected to generate a coefficient wave with the correct number of
//	parameters and return this wave to the s.cw parameter. Additionally, the s.holdStr can be set
//	to hold one parameter (usually the baseline) if the option is activated in the menu.
// 
//	Additional custom fit functions are provided with the package:
//		- Fermi-Dirac function with linear slope
//		- Line with x-crossing as input parameter (instead of y-crossing)
//		- Angular Distribution function (e.g., for Photoelectron Angular Distributions)
//		- Double Gauss function
//		- Exponentially Modified Gaussian (ExpModGauss)
//		- Convolution of two exponential decays (ExpConvExp)
//
//	Support for standard fit functions with the extended options is provided as well.
//
// 2022-04-07 - ver. 1.00:	Initial public release.
// 2022-04-08 - ver. 1.01:	Added textbox preferences.
// 2022-05-11 - ver. 1.02:	Fixed bug: Missing M_Covar for Line fit was not handled properly.
// 2022-07-14 - ver. 1.03:	Improved guess for EF parameter in Fermi-Line guess function.
//							Improved derived parameters for EMG peak shape.
// 2022-10-18 - ver. 1.04:	'Clean-up' now only removes fit-related variables.
//							Derived values will be added to wave notes as well.
//							Improved guess for DoubleGauss function.
// 2022-10-27 - ver. 1.05:	Added ExpConvExp function (thanks to Oleg Dyachok from Uppsala University for the idea).
//							Now negative EMG peaks can be fitted.
//							Fixed bug: 'use cursor' was not working for small delta-scaling.
// 2022-11-11 - ver. 1.06:	Fixed bug: EMG and ExpConvExp guesses did not work for XY data when the y-wave was scaled.
//							Added peak-area output to EMG peak.
//							Added option to align the fit trace to the source trace (thanks to Oleg Dyachok from Uppsala University for the idea).
// 2022-11-27 - ver. 1.07:	Added support for fitting traces from 2D waves.
//							Added additional derived parameters for ExpConvExp (thanks to Oleg Dyachok for the help).
//							Added confidence intervals to the area of the ExpModGauss and ExpConvExp functions.
//							Changed behavior: Now the cursor placement does not set the baseline, but only decides the fit range.
// 2022-12-07 - ver. 1.08:	Fixed bug: Derived parameters for ExpConvExp were wrong.
//							Now SuperQuickFit() can be invoked in user code and the command line.
//							Fixed bug: 2D data could not be properly read if the name was liberal.
// 2023-04-06 - ver. 1.10:	Added option to create a separate fit wave for each column of a 2D wave.
//							Traces are appended to the correct axis if the source is not plotted on the default axis.
//							New settings menu to choose color and style for the fit trace.
//							Added support for contour plots.
//							Global preferences are not overwritten completely by local references anymore.
//							Rearranged entries in the settings menu.
// 2023-11-03 - ver. 1.11:	Improved initial guesses for the DoubleGauss peak coefficient.
//							Added support for data plotted with column tags.
//							Added support for x-wave columns (traces plotted against a column of a 2D wave).
//							Fixed bug: It was not possible to set the start and end points with cursors directly.
//							Now the proper axis of the image / trace is used.
//							Column tags, instead of column numbers, are used for CurveFit commands if available.
//________________________________________________________________________________________________

strConstant QuickFit_DefaultFuncs1D = "Gauss;Lor;Voigt;Sin;Line;-;HillEquation;Sigmoid;Power;LogNormal;Log;-;Exp;Exp_xOffset;DblExp;DblExp_Peak;DblExp_xOffset;"
strConstant QuickFit_DefaultFuncs2D = "Gauss2D;"

static strConstant CoefGen_SUFFIX = "_prepareCoef"												// name suffixes for coef and output functions 
static strConstant derivedVal_SUFFIX = "_derivedVals"
static strConstant localSettingsPath = "root:QuickFitMenu_settings"								// where local settings are saved

function dummyFunc_prepareCoef(STRUCT SuperQuickFitStruct &s)									// dummy functions for coef and output
	return -1
end
function/S dummyFunc_derivedVals(STRUCT SuperQuickFitStruct &s)
	return ""
end

Structure SuperQuickFitStruct
	wave cw			// coef wave
	wave sw			// covariance matrix (M_Covar)
	wave data		// data wave (1D or 2D)
	wave xw			// x wave
	wave yw			// y wave (for images; not supported yet)
	int32 pntMin	// start point of fit range
	int32 pntMax	// end point of fit range
	int32 pntMinY	// start point in y direction (for images)
	int32 pntMaxY	// end point  in y direction (for images)
	int32 pntCol	// column to use for traces from 2D wave
	double xBase	// x baseline / start offset
	double yBase	// y baseline / start offset
	int16 holdMode	// current hold setting: [0] = no hold, [1] = force o zero, [2] = force to min
	int16 plotFull	// setting to plot over full range
	int16 doTextbox	// setting for text-box drawing
	string holdStr	// hold string for the fit
EndStructure

// +++++++++++++++++++++++++++++ persistent preferences support ++++++++++++++++++++++++++++++++++

static strConstant defaultSettings	= "offsetMode:0;plotFull:0;plotAligned:0;fitRangeMode:1;addTextBox:0;weightError:1;outCleanup:0;textboxOpt:791;multiColMode:0;trStyleMode:0;trLineSize:1;trColorMode:0;trColor:65535,0,0,65535;"
static strConstant kPackageName		= "SuperQuickFit"
static strConstant kPrefsFileName	= "SuperQuickFitPrefs.bin"
static Constant kPrefsVersion		= 100	// version 1.00

Structure SuperQuickFitPrefs
	uint32	version					// Preferences structure version number. 100 means 1.00.
	int16	offsetMode
	int16	plotFull
	int16	fitRangeMode
	int16	addTextBox
	int16	weightError
	int16	outCleanup
	int16	textboxOpt
	int16	plotAligned				// new since version 1.06
	int16	trStyleMode				// version 1.09
	int16	trColorMode				// version 1.09
	double	trLineSize				// version 1.09
	uint16	trColor[4]				// version 1.09
	int16	multiColMode			// version 1.09
	char reserved[200-4*1-15*2-8*1]	// Reserved for future use
EndStructure

static function/S LoadPrefs()
	STRUCT SuperQuickFitPrefs s
	LoadPackagePreferences kPackageName, kPrefsFileName, 0, s
	string settings = ""
	if (V_flag!=0 || V_bytesRead==0 || s.version!=kPrefsVersion)								// If error or prefs not found or not valid, initialize them.
		AnalysisTools#SavePrefs(defaultSettings)
	endif
	sPrintf settings, "offsetMode:%d;plotFull:%d;plotAligned:%d;fitRangeMode:%d;addTextBox:%d;weightError:%d;outCleanup:%d;textboxOpt:%d;", s.offsetMode, s.plotFull, s.plotAligned, s.fitRangeMode, s.addTextBox, s.weightError, s.outCleanup, s.textboxOpt
	sPrintf settings, "%smultiColMode:%d;trStyleMode:%d;trLineSize:%d;trColorMode:%d;trColor:%d,%d,%d,%d;", settings, s.multiColMode, s.trStyleMode, s.trLineSize, s.trColorMode, s.trColor[0], s.trColor[1], s.trColor[2], s.trColor[3]
	return settings
end

static function SavePrefs(string settings)
	STRUCT SuperQuickFitPrefs s
	s.version		= kPrefsVersion
	s.offsetMode	= NumberByKey("offsetMode" ,settings)
	s.plotFull		= NumberByKey("plotFull" ,settings)
	s.fitRangeMode	= NumberByKey("fitRangeMode" ,settings)
	s.addTextBox	= NumberByKey("addTextBox" ,settings)
	s.weightError	= NumberByKey("weightError" ,settings)
	s.outCleanup	= NumberByKey("outCleanup" ,settings)
	s.textboxOpt	= NumberByKey("textboxOpt" ,settings)
	s.plotAligned	= NumberByKey("plotAligned" ,settings)
	s.trStyleMode	= NumberByKey("trStyleMode" ,settings)
	s.trLineSize	= NumberByKey("trLineSize" ,settings)
	s.trColorMode	= NumberByKey("trColorMode" ,settings)
	s.multiColMode	= NumberByKey("multiColMode" ,settings)
	string colors	= StringByKey("trColor" ,settings)
	s.trColor[0]	= str2num(StringFromList(0, colors, ","))
	s.trColor[1]	= str2num(StringFromList(1, colors, ","))
	s.trColor[2]	= str2num(StringFromList(2, colors, ","))
	s.trColor[3]	= str2num(StringFromList(3, colors, ","))
	SavePackagePreferences kPackageName, kPrefsFileName, 0, s
	return V_flag
end

//################################################################################################

Menu "GraphPopup", dynamic
	SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Super Quick Fit")
		SubMenu "Settings"
			AnalysisTools#QuickFitMenuItem("useCursor"),	/Q, AnalysisTools#QuickFitMenuSettings("useCursor")
			AnalysisTools#QuickFitMenuItem("useAxRange"),	/Q, AnalysisTools#QuickFitMenuSettings("useAxRange")
			AnalysisTools#QuickFitMenuItem("multiColMode"),	/Q, AnalysisTools#QuickFitMenuSettings("multiColMode")
			"Fit Trace Preferences ...",					/Q, AnalysisTools#QickFitOutputTraceOptions()
			"-"
			AnalysisTools#QuickFitMenuItem("zeroBase"),		/Q, AnalysisTools#QuickFitMenuSettings("zeroBase")
			AnalysisTools#QuickFitMenuItem("fixedBase"),	/Q, AnalysisTools#QuickFitMenuSettings("fixedBase")
			AnalysisTools#QuickFitMenuItem("weightError"),	/Q, AnalysisTools#QuickFitMenuSettings("weightError")
			"-"
			AnalysisTools#QuickFitMenuItem("outCleanup"),	/Q, AnalysisTools#QuickFitMenuSettings("outCleanup")
			AnalysisTools#QuickFitMenuItem("plotFull"),		/Q, AnalysisTools#QuickFitMenuSettings("plotFull")
			AnalysisTools#QuickFitMenuItem("plotAligned"),	/Q, AnalysisTools#QuickFitMenuSettings("plotAligned")
			AnalysisTools#QuickFitMenuItem("addTextBox"),	/Q, AnalysisTools#QuickFitMenuSettings("addTextBox")
			"\t» Textbox Preferences ...",					/Q, AnalysisTools#QickFitTextBoxOptions()
			"-"
			"Clear Local Settings",							/Q, KillStrings/Z $localSettingsPath
		End
		"-"
		AnalysisTools#QuickFitMenuUserEntries(), /Q, AnalysisTools#SuperQuickFit_Wrapper("")
		"-"
		SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Poly")
			"1;2;3;4;5;6;7;8;9;10;", /Q, AnalysisTools#SuperQuickFit_Wrapper("poly")
		End
		SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Poly_xOffset")
			"1;2;3;4;5;6;7;8;9;10;", /Q, AnalysisTools#SuperQuickFit_Wrapper("poly_XOffset")
		End
		SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Poly2D")
			"1;2;3;4;5;6;7;8;9;10;", /Q, AnalysisTools#SuperQuickFit_Wrapper("Poly2D")
		End
		"-"
		AnalysisTools#QuickFitMenuDefaultEntries(), /Q, AnalysisTools#SuperQuickFit_Wrapper("")
	End
End
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Menu "TracePopup", dynamic
	SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Super Quick Fit")
		SubMenu "Settings"
			AnalysisTools#QuickFitMenuItem("useCursor"),	/Q, AnalysisTools#QuickFitMenuSettings("useCursor")
			AnalysisTools#QuickFitMenuItem("useAxRange"),	/Q, AnalysisTools#QuickFitMenuSettings("useAxRange")
			AnalysisTools#QuickFitMenuItem("multiColMode"),	/Q, AnalysisTools#QuickFitMenuSettings("multiColMode")
			"Fit Trace Preferences ...",					/Q, AnalysisTools#QickFitOutputTraceOptions()
			"-"
			AnalysisTools#QuickFitMenuItem("zeroBase"),		/Q, AnalysisTools#QuickFitMenuSettings("zeroBase")
			AnalysisTools#QuickFitMenuItem("fixedBase"),	/Q, AnalysisTools#QuickFitMenuSettings("fixedBase")
			AnalysisTools#QuickFitMenuItem("weightError"),	/Q, AnalysisTools#QuickFitMenuSettings("weightError")
			"-"
			AnalysisTools#QuickFitMenuItem("outCleanup"),	/Q, AnalysisTools#QuickFitMenuSettings("outCleanup")
			AnalysisTools#QuickFitMenuItem("plotFull"),		/Q, AnalysisTools#QuickFitMenuSettings("plotFull")
			AnalysisTools#QuickFitMenuItem("plotAligned"),	/Q, AnalysisTools#QuickFitMenuSettings("plotAligned")
			AnalysisTools#QuickFitMenuItem("addTextBox"),	/Q, AnalysisTools#QuickFitMenuSettings("addTextBox")
			"\t» Textbox Preferences ...",					/Q, AnalysisTools#QickFitTextBoxOptions()
			"-"
			"Clear Local Settings",							/Q, KillStrings/Z $localSettingsPath
		End
		"-"
		AnalysisTools#QuickFitMenuUserEntries(), /Q, AnalysisTools#SuperQuickFit_Wrapper("")
		"-"
		SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Poly")
			"1;2;3;4;5;6;7;8;9;10;", /Q, AnalysisTools#SuperQuickFit_Wrapper("poly")
		End
		SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Poly_xOffset")
			"1;2;3;4;5;6;7;8;9;10;", /Q, AnalysisTools#SuperQuickFit_Wrapper("poly_XOffset")
		End
		SubMenu AnalysisTools#QuickFitMenuSubMenuEntries("Poly2D")
			"1;2;3;4;5;6;7;8;9;10;", /Q, AnalysisTools#SuperQuickFit_Wrapper("Poly2D")
		End
		"-"
		AnalysisTools#QuickFitMenuDefaultEntries(), /Q, AnalysisTools#SuperQuickFit_Wrapper("")
	End
End

//################################## SQF menu interaction ########################################

static function SuperQuickFit_Wrapper(string option)
	GetLastUserMenuInfo																			// figure out graph and trace names
	string fitFunc = SelectString(StringMatch(option, "poly*"),"",option+" ") + S_value			// grab fitFunc from selected item: for poly fits append degree n
	SuperQuickFit(S_graphName, S_traceName, fitFunc)
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S QuickFitMenuSubMenuEntries(string entry)										// sub menu entry generator
	int noTraces = 0, notFor2D = 0; [noTraces, notFor2D] = AnalysisTools#checkTraceOrImage()
	if (CmpStr(entry, "Super Quick Fit") == 0)													// SQF main menu
		return SelectString(noTraces && notFor2D,"","(")+entry									// nothing useable found
	else
		return SelectString((StringMatch(entry,"*2D") && !notFor2D) %^ noTraces,"","(")+entry	// either both no traces and 2D-compatible (1 %^ 1) or both traces and 2D-incompatible (0 %^ 0)
	endif
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S QuickFitMenuUserEntries()														// entries from user functions
	string funcList = RemoveFromList("dummyFunc",ReplaceString(CoefGen_SUFFIX,FunctionList("*"+CoefGen_SUFFIX, ";", ""),""))
	int noTraces = 0, notFor2D = 0; [noTraces, notFor2D] = AnalysisTools#checkTraceOrImage()
	return AnalysisTools#deactivateList(funcList, noTraces)
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S QuickFitMenuDefaultEntries()													// default fit list
	int noTraces = 0, notFor2D = 0; [noTraces, notFor2D] = AnalysisTools#checkTraceOrImage()
	return AnalysisTools#deactivateList(QuickFit_DefaultFuncs2D, !noTraces || notFor2D) + AnalysisTools#deactivateList(QuickFit_DefaultFuncs1D, noTraces)
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S getTopGraph()
	return StringFromList(0,WinList("*", ";", "WIN:1"))
end

static function [int noTraces, int notFor2D] checkTraceOrImage()								// checks for compatible image or trace in top graph
	string gName = getTopGraph()
	if (!strlen(gName))
		return [1, 1]
	endif
	string image = ImageInfo(gName,"",0)
	noTraces = ItemsInList(TraceNameList(gName,";",1)) == 0
	if (noTraces)
		if (strlen(image))
			wave/Z imgXW = $(StringByKey("XWAVEDF",image)+StringByKey("XWAVE",image))
			wave/Z imgYW = $(StringByKey("YWAVEDF",image)+StringByKey("YWAVE",image))									
			notFor2D = WaveExists(imgXW) || WaveExists(imgYW)									// no support for x/y waves with images yet
		else
			notFor2D = ItemsInList(ContourNameList(gName,";")) == 0								// otherwise look for contour plots
		endif
	endif
	return [noTraces, notFor2D]
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S deactivateList(string list, int deactivate)									// deactivate menu entry list (prepend '(' character)
	if (deactivate)
		return "("+RemoveEnding(ReplaceString(";", list, ";("),"(")
	else
		return list
	endif
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S QuickFitMenuItem(string which)												// returns menu entry for settings
	string settings = StrVarOrDefault(localSettingsPath,AnalysisTools#LoadPrefs())
	StrSwitch(which)
		case "zeroBase":
			return SelectString(NumberByKey("offsetMode"	,settings) == 1,"","!"+num2char(18)) + "Offset: Force to Zero"
		case "fixedBase":
			return SelectString(NumberByKey("offsetMode"	,settings) == 2,"","!"+num2char(18)) + "Offset: Force to Minimum"
		case "plotFull":
			return SelectString(NumberByKey("plotFull"		,settings) == 1,"","!"+num2char(18)) + "Result: Plot Full Range"
		case "plotAligned":
			return SelectString(NumberByKey("plotAligned"	,settings) == 1,"","!"+num2char(18)) + "Result: Align Fit with Trace"
		case "addTextBox":
			return SelectString(NumberByKey("addTextBox"	,settings) == 1,"","!"+num2char(18)) + "Result: Add TextBox"
		case "multiColMode":
			return SelectString(NumberByKey("multiColMode"	,settings) == 1,"","!"+num2char(18)) + "Fit Individual 2D Columns"
		case "useCursor":
			return SelectString(NumberByKey("fitRangeMode"	,settings) == 1,"","!"+num2char(18)) + "Fit Between Cursors"
		case "useAxRange":
			return SelectString(NumberByKey("fitRangeMode"	,settings) == 2,"","!"+num2char(18)) + "Fit Within Axis Range"
		case "weightError":
			return SelectString(NumberByKey("weightError"	,settings) == 1,"","!"+num2char(18)) + "Weight from Error Bar Wave"
		case "outCleanup":
			return SelectString(NumberByKey("outCleanup"	,settings) == 1,"","!"+num2char(18)) + "Clean up Output Data"
	EndSwitch
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function QuickFitMenuSettings(string which)												// write settings from selection
	string globalSettings = AnalysisTools#LoadPrefs()											// saves global (PackagePrefs) and local (string) settings separately
	string localSettings  = StrVarOrDefault(localSettingsPath,globalSettings)
	string/G $localSettingsPath  = AnalysisTools#SQF_doGeneralSettings(which, localSettings)
	return AnalysisTools#SavePrefs(AnalysisTools#SQF_doGeneralSettings(which, globalSettings))
end

static function/S SQF_doGeneralSettings(string which, string settings)							// replace selected setting parameter
	StrSwitch(which)
		case "zeroBase":
			return ReplaceNumberByKey("offsetMode"	,settings ,(NumberByKey("offsetMode"	,settings) == 1 ? 0 : 1))
		case "fixedBase":
			return ReplaceNumberByKey("offsetMode"	,settings ,(NumberByKey("offsetMode"	,settings) == 2 ? 0 : 2))
		case "useCursor":
			return ReplaceNumberByKey("fitRangeMode",settings ,(NumberByKey("fitRangeMode"	,settings) == 1 ? 0 : 1))
		case "useAxRange":
			return ReplaceNumberByKey("fitRangeMode",settings ,(NumberByKey("fitRangeMode"	,settings) == 2 ? 0 : 2))
		default:
			return ReplaceNumberByKey(which			,settings ,(NumberByKey(which 			,settings) == 1 ? 0 : 1))
	EndSwitch
end

//##################################### Text Options Panel #######################################

static StrConstant kOptBits = "Title;Date;Time;FitType;FitFunc;ModelWave;YWave;XWave;CoefValues;CoefErrors;"

static function QickFitTextBoxOptions()
	if (WinType("SQF_TextBoxOpt"))
		return 0
	endif
	string optTitle = "Select All;Title;Date and Time;Date;Time;Fit Type;Fit Function Name;Waves;Model Wave;Y Wave;X Wave(s);Coefficient Report;Coefficient Values;Coefficient Errors;"
	string optNames = "All;Title;DateTime;Date;Time;FitType;FitFunc;Waves;ModelWave;YWave;XWave;CoefReport;CoefValues;CoefErrors;"
	string optAlign = "0;0;0;15;15;0;0;0;15;15;15;0;15;15;"
	GetMouse
	NewPanel/FLT=2/K=1/W=(V_left,V_top,V_left+155,V_top+400)/N=SQF_TextBoxOpt as "Textbox Preferences"
	int i, dist = 25
	for (i=0; i<ItemsInList(optTitle); i++)
		CheckBox $("c"+StringFromList(i,optNames)) ,pos={15+str2num(StringFromList(i,optAlign)),15+dist*i} ,size={100,23} ,title=StringFromList(i,optTitle) ,proc=AnalysisTools#SQF_TextboxOptions
	endfor
	Button b_Save	,pos={10 ,15+dist*i} ,size={60,22} ,title="Save"	,proc=AnalysisTools#SQF_TextBoxButton
	Button b_Cancel	,pos={85 ,15+dist*i} ,size={60,22} ,title="Cancel"	,proc=AnalysisTools#SQF_TextBoxButton
	AnalysisTools#SQF_TextBoxLoadFromSettings()
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function SQF_TextboxOptions(STRUCT WMCheckboxAction &s) : CheckBoxControl
	if(s.eventCode == 2)
		string ctrlList = ""
		StrSwitch(s.ctrlName)
			case "cDateTime":
				ctrlList = "cDate;cTime;"
			break
			case "cWaves":
				ctrlList = ControlNameList(s.win, ";", "*Wave")
			break
			case "cCoefReport":
				ctrlList = ControlNameList(s.win, ";", "cCoef*")
			break
			case "cAll":
				ctrlList = ControlNameList(s.win, ";", "c*")
			break
		EndSwitch
		int i, count = 0
		for (i=0; i<ItemsInList(ctrlList); i++)
			CheckBox $(StringFromList(i,ctrlList)) ,win=$s.win ,value=s.checked
		endfor
		ctrlList = RemoveFromList("cAll", ControlNameList(s.win, ";", "c*"))					// now see if the Check All box needs to be toggled
		for (i=0; i<ItemsInList(ctrlList); i++)
			ControlInfo/W=$s.win $("c"+StringFromList(i,kOptBits)); count += V_value
		endfor
		CheckBox cAll ,win=$s.win ,value=(ItemsInList(ctrlList) == count)
	endif
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function SQF_TextBoxButton(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode == 2)
		if(CmpStr(s.ctrlName,"b_Save") == 0)
			AnalysisTools#SQF_TextBoxWriteSettings()											// save options
		endif
		KillWindow $s.win
	endif
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function SQF_TextBoxWriteSettings()														// save textbox bit settings
	variable i, textboxOpt = 0
	for (i=0; i<ItemsInList(kOptBits); i++)														// read checked bits from checkboxes
		ControlInfo/W=SQF_TextBoxOpt $("c"+StringFromList(i,kOptBits))
		textboxOpt = V_Value ? textboxOpt | (2^i) : textboxOpt
	endfor
	string globalSettings = AnalysisTools#LoadPrefs()
	string localSettings  = StrVarOrDefault(localSettingsPath,globalSettings)
	string/G $localSettingsPath  = ReplaceNumberByKey("textboxOpt",localSettings,textboxOpt)	// write new option
	return AnalysisTools#SavePrefs(ReplaceNumberByKey("textboxOpt",globalSettings,textboxOpt))
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function SQF_TextBoxLoadFromSettings()													// transfers checked bits to checkboxes
	string settings = StrVarOrDefault(localSettingsPath,AnalysisTools#LoadPrefs())
	variable i, txtBits = NumberByKey("textboxOpt",settings)
	for (i=0; i<ItemsInList(kOptBits); i++)
		CheckBox $("c"+StringFromList(i,kOptBits)) ,win=SQF_TextBoxOpt ,value=((txtBits & 2^i) != 0)
	endfor
	CheckBox cDateTime	 ,win=SQF_TextBoxOpt ,value=((txtBits & 2^1)!=0 && (txtBits & 2^2)!=0)
	CheckBox cWaves		 ,win=SQF_TextBoxOpt ,value=((txtBits & 2^5)!=0 && (txtBits & 2^6)!=0 && (txtBits & 2^7)!=0)
	CheckBox cCoefReport ,win=SQF_TextBoxOpt ,value=((txtBits & 2^8)!=0 && (txtBits & 2^9)!=0)
	CheckBox cAll		 ,win=SQF_TextBoxOpt ,value=( txtBits == 1023)							// all 10 bits are set
	return 0
end

//################################### Color Options Panel ########################################

static StrConstant kTraceColorOpt = "Default Color;Single Fixed Color;Same Color as Input;Complementary to Input;Lower Saturation than Input;"

static function QickFitOutputTraceOptions()
	if (WinType("SQF_FitTraceOpt"))
		return 0
	endif
	GetMouse
	NewPanel/FLT=2/K=1/W=(V_left,V_top,V_left+300,V_top+100)/N=SQF_FitTraceOpt as "Fit Trace Preferences"
	int i=0, dist = 28
	PopupMenu p_TrColor		,pos={235,	10+dist*i}		,size={45,23}	,title=""				,bodyWidth=45	,value="*COLORPOP*"
	PopupMenu p_TrColorMode	,pos={15,	10+dist*i++}	,size={265,23}	,title="Color Mode:"	,bodyWidth=195	,value=#("\""+kTraceColorOpt+"\"")
	PopupMenu p_TrStyle		,pos={15,	10+dist*i}		,size={149,23}	,title="Line Style:"	,bodyWidth=80	,value="*LINESTYLEPOP*"
	SetVariable v_TrLSize	,pos={175,	10+dist*i++}	,size={105,23}	,title="Line Size:"		,bodyWidth=50	,value=_NUM:1	,limits={0,10,1}
	Button b_Save			,pos={50,	10+dist*i} 		,size={80,22}	,title="Save"
	Button b_Cancel			,pos={170,	10+dist*i}		,size={80,22}	,title="Cancel"
	ModifyControlList ControlNameList("SQF_FitTraceOpt",";","b_*") proc=AnalysisTools#SQF_FitOptionsButton
	ModifyControlList ControlNameList("SQF_FitTraceOpt",";","p_*") proc=AnalysisTools#SQF_FitOptionsPopup
	AnalysisTools#SQF_FitTraceLoadFromSettings()
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function SQF_FitOptionsButton(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode == 2)
		if(CmpStr(s.ctrlName,"b_Save") == 0)
			AnalysisTools#SQF_FitTraceWriteSettings()											// save options
		endif
		KillWindow $s.win
	endif
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

function SQF_FitOptionsPopup(STRUCT WMPopupAction &s) : PopupMenuControl
	if (s.eventCode == 2 && CmpStr(s.ctrlName,"p_TrColorMode") == 0)
		int off = !StringMatch(s.popStr,"Single Fixed Color")
		PopupMenu p_TrColor 	,win=$s.win ,disable=off
		PopupMenu p_TrColorMode ,win=$s.win ,size={205+60*off,23} ,bodyWidth=135+60*off
	endif
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function SQF_FitTraceWriteSettings()
	string globalSettings = AnalysisTools#LoadPrefs()
	string localSettings  = StrVarOrDefault(localSettingsPath,globalSettings)
	string/G $localSettingsPath  = AnalysisTools#SQF_doTraceStyleSettings(localSettings)
	return AnalysisTools#SavePrefs(AnalysisTools#SQF_doTraceStyleSettings(globalSettings))
end

static function/S SQF_doTraceStyleSettings(string settings)										// replace fit trace settings
	string color = ""
	ControlInfo/W=SQF_FitTraceOpt v_TrLSize;		settings = ReplaceNumberByKey("trLineSize",settings,V_Value)
	ControlInfo/W=SQF_FitTraceOpt p_TrStyle;		settings = ReplaceNumberByKey("trStyleMode",settings,V_Value-1)
	ControlInfo/W=SQF_FitTraceOpt p_TrColorMode;	settings = ReplaceNumberByKey("trColorMode",settings,V_Value-1)
	ControlInfo/W=SQF_FitTraceOpt p_TrColor;		sPrintf color, "%d,%d,%d,%d", V_Red, V_Green, V_Blue, V_Alpha
	return ReplaceStringByKey("trColor",settings,color)
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function SQF_FitTraceLoadFromSettings()
	string settings = StrVarOrDefault(localSettingsPath,AnalysisTools#LoadPrefs())
	int FitColor	= NumberByKey("trColorMode",settings);	FitColor = FitColor <  0 ? 0 : FitColor
	int FitStyle	= NumberByKey("trStyleMode",settings);	FitStyle = FitStyle <  0 ? 0 : FitStyle
	int FitLSize	= NumberByKey("trLineSize",settings);	FitLSize = FitLSize <= 0 ? 1 : limit(FitLSize,0,10)
	string color = StringByKey("trColor",settings)
	int A = str2num(StringFromList(2,color,","))
	int R = str2num(StringFromList(0,color,",")); R = R < 0 || A <= 0 ? 65535 : R
	int G = str2num(StringFromList(1,color,",")); G = G < 0 || A <= 0 ? 0 : G
	int B = str2num(StringFromList(2,color,",")); B = B < 0 || A <= 0 ? 0 : B
	A = A <= 0 ? 65535 : A
	int expand = 60*(FitColor!=1)
	PopupMenu p_TrColorMode	,win=SQF_FitTraceOpt	,mode=FitColor+1	,size={205+expand,23}	,bodyWidth=135+expand
	PopupMenu p_TrStyle		,win=SQF_FitTraceOpt	,mode=FitStyle+1
	PopupMenu p_TrColor		,win=SQF_FitTraceOpt	,popColor=(R,G,B,A)	,disable=(FitColor!=1)
	SetVariable v_TrLSize	,win=SQF_FitTraceOpt	,value=_NUM:FitLSize
	return 0
end

//________________________________________________________________________________________________
//									Super Quick Fit Kernel
//________________________________________________________________________________________________

function SuperQuickFit(string gName, string trace, string fitFunc)
	STRUCT SuperQuickFitStruct s
	
	int sepPos = strsearch(trace,"#",0);	sepPos = sepPos == -1 ? strlen(trace) : sepPos-1
	trace = PossiblyQuoteName(trace[0,sepPos])+trace[sepPos+1,inf]								// make sure the trace name has the correct quotes
	
	// +++++++++++++++++++++++++++++ first some error handling +++++++++++++++++++++++++++++++++++
	if (!strlen(gName))
		gName = getTopGraph()																	// try to get a valid graph
		if (!strlen(gName))
			Print "SQF error: Graph not found."
			return -2
		endif
	endif
	
	if (WinType(gName) != 1)
		Print "SQF error: Graph '"+gName+"' does not exist."
		return -3
	endif
	
	string traceList = TraceNameList(gName,";",1), contourList = TraceNameList(gName,";",2)
	if (strlen(trace) && FindListItem(trace, traceList+contourList,";",0,0) < 0)
		Print "SQF error: Trace "+trace+" not in graph "+gName+"."
		return -4
	endif
	
	string fitType = ""
	if (FindListItem(fitFunc,RemoveFromList("-",QuickFit_DefaultFuncs1D+QuickFit_DefaultFuncs2D),";",0,0) > -1)	// is a default function
		fitType = "CurveFit"
	elseif (FindListItem(fitFunc, FunctionList("*",";","KIND:2,NINDVARS:2"),";",0,0) > -1)						// is probably a user function
		fitType = "FuncFit"
	else
		if (StringMatch(fitFunc, "poly *") || StringMatch(fitFunc, "poly_xOffset *") || StringMatch(fitFunc, "poly2D *"))	// for poly fits
			fitType = "CurveFit"
			fitFunc = RemoveEnding(fitFunc, ",") + ","											// make sure there is a comma at the end
		else
			Print "SQF error: Not a valid function type."
			return -5																			// not compatible
		endif
	endif
	
	variable isContour = 0
	if (FindListItem(trace, contourList,";",0,0) > -1)											// contour plot
		sepPos = strsearch(trace,"=",0);	sepPos = sepPos == -1 ? strlen(trace) : sepPos-1
		wave/Z s.data = ContourNameToWaveRef(gName, ReplaceString("'",trace[0,sepPos],""))
		if (!WaveExists(s.data))
			Print "SQF error: Not a valid contour trace."
			return -6
		endif
		isContour = 1
	endif
	
	string image = ""
	if (!strlen(trace))																			// was called from graph popup => find top trace
		image = ImageInfo(gName,"",0)															// might also be an image plot (takes preference)
		trace = StringFromList(0, traceList)
		if (!strlen(trace))
			wave/Z s.data = ContourNameToWaveRef(gName, StringFromList(0,ContourNameList(gName,";")))
			isContour = WaveExists(s.data)
		endif
	endif
	
	if (!strlen(trace) && !strlen(image) && !isContour || !strlen(fitFunc) || !strlen(fitType))	// still no trace, image or function => abort
		Print "SQF error: No valid target or function."
		return -1
	endif
	
	// ++++++++++++++++++++++ set up the input data and load options ++++++++++++++++++++++++++++
	variable isImage = strlen(image), col2D_yw = NaN, col2D_xw = NaN
	string Xcol_label = "", Ycol_label = ""
	if (isImage)
		wave s.data = $(StringByKey("ZWAVEDF",image)+PossiblyQuoteName(StringByKey("ZWAVE",image)))
		wave/Z s.xw = $(StringByKey("XWAVEDF",image)+PossiblyQuoteName(StringByKey("XWAVE",image)))
		wave/Z s.yw = $(StringByKey("YWAVEDF",image)+PossiblyQuoteName(StringByKey("YWAVE",image)))
		if (WaveExists(s.xw) || WaveExists(s.yw))												// no support for x/y waves with images yet
			return -100
		endif
		s.pntMinY = 0
		s.pntMaxY = DimSize(s.data,1)-1
	elseif (isContour)
		s.pntMinY = 0
		s.pntMaxY = DimSize(s.data,1)-1
	else
		wave s.data = TraceNameToWaveRef(gName, trace)
		wave/Z s.xw = XWaveRefFromTrace(gName, trace)
		col2D_yw = getCol(gName, trace, 0, Ycol_label)											// will be appended to the fit command
		col2D_xw = getCol(gName, trace, 1, Xcol_label)
	endif
	
	int isTraceOf2D = numtype(col2D_yw) == 0
	int isXTraceCol = numtype(col2D_xw) == 0
	int xCol = isXTraceCol ? col2D_xw : 0
	s.pntCol = isTraceOf2D ? col2D_yw : 0
	s.pntMin = 0
	s.pntMax = DimSize(s.data,0)-1
	
	string settings = StrVarOrDefault(localSettingsPath,AnalysisTools#LoadPrefs())				// load settings
	string/G $localSettingsPath = settings														// save used settings to experiment
	int textboxOpt	= NumberByKey("textboxOpt",settings)
	int rangeMode	= NumberByKey("fitRangeMode",settings)
	int doWeight	= NumberByKey("weightError",settings)
	int doCleanUp	= NumberByKey("outCleanup",settings)
	
	int doFitAlign	= NumberByKey("plotAligned",settings);	doFitAlign = doFitAlign < 0 ? 0 : doFitAlign		// these option were introduced later and may not be initialized yet
	int doMultiOut	= NumberByKey("multiColMode",settings);	doMultiOut = doMultiOut < 0 ? 0 : doMultiOut
	int setFitColor	= NumberByKey("trColorMode",settings);	setFitColor = setFitColor < 0 ? 0 : setFitColor
	int setFitStyle	= NumberByKey("trStyleMode",settings);	setFitStyle = setFitStyle < 0 ? 0 : setFitStyle
	int FitLineSize	= NumberByKey("trLineSize",settings);	FitLineSize = FitLineSize <= 0 ? 1 : limit(FitLineSize,0,10)
	string FitColor = StringByKey("trColor",settings)
	
	s.plotFull	= NumberByKey("plotFull",settings)
	s.doTextbox = NumberByKey("addTextBox",settings)
	s.holdMode  = NumberByKey("offsetMode",settings)
	s.holdStr 	= "1"																			// default is baseline hold
	
	// +++++++++++++++++++++++++ find fit range and baseline values ++++++++++++++++++++++++++++++	
	int i = 0
	string csrA = "", csrB = "", csrList = "A;B;C;D;E;F;G;H;I;J;"
	Make/Free/D/N=4 csrA_vals = NaN, csrB_vals = NaN
	for (; i < ItemsInList(csrList);)															// see which cursor is placed
		csrA = StringFromList(i++,csrList)
		if (strlen(CsrInfo($csrA,gName)))
			csrA_vals = {xcsr($csrA, gName), pcsr($csrA, gName), hcsr($csrA, gName), vcsr($csrA, gName)}
			break
		endif
	endfor
	for (; i < ItemsInList(csrList);)															// poll the second cursor
		csrB = StringFromList(i++,csrList)
		if (strlen(CsrInfo($csrB,gName)))
			csrB_vals = {xcsr($csrB, gName), pcsr($csrB, gName), hcsr($csrB, gName), vcsr($csrB, gName)}
			break
		endif
	endfor
	
	variable xMin = NaN, xMax = NaN, yMin = NaN, yMax = NaN, temp
	if (rangeMode == 1)
		if (!numtype(csrA_vals[0]))																// not a free cursor
			if (WaveRefsEqual(s.data, CsrWaveRef($csrA,gName)))									// cursor is attached to the same wave -> use points directly
				s.pntMin = csrA_vals[1]
			else
				xMin = csrA_vals[0]
			endif
		else
			xMin = csrA_vals[2]
		endif
		if (!numtype(csrB_vals[0]))
			if (WaveRefsEqual(s.data, CsrWaveRef($csrB,gName)))
				s.pntMax = csrB_vals[1]
			else
				xMax = csrB_vals[0]
			endif
		else
			xMax = csrB_vals[2]
		endif
		if (s.pntMin == s.pntMax && numtype(xMin) && numtype(xMax))								// cursors at the same position => undo
			s.pntMin = 0
			s.pntMax = DimSize(s.data,0)-1
		endif
	elseif (rangeMode == 2)
		if (isImage)
			GetAxis/W=$gName/Q $StringByKey("XAXIS",image)
		else
			GetAxis/W=$gName/Q $StringByKey("XAXIS",TraceInfo(gName,trace,0))
		endif
		xMin = !V_flag ? V_min : xMin
		xMax = !V_flag ? V_max : xMax
	endif
	if (!numtype(xMin) && !numtype(xMax) && xMin != xMax)
		if (WaveExists(s.xw))
			Duplicate/Free/R=[][xCol] s.xw, temp_xw
			FindLevel/P/Q temp_xw, xMin; s.pntMin = (!v_flag && v_levelX > 0 && v_levelX < DimSize(s.data,0)-1) ? v_levelX : s.pntMin
			FindLevel/P/Q temp_xw, xMax; s.pntMax = (!v_flag && v_levelX > 0 && v_levelX < DimSize(s.data,0)-1) ? v_levelX : s.pntMax
		else
			s.pntMin = limit(x2pnt(s.data,xMin), 0, DimSize(s.data,0)-1)
			s.pntMax = limit(x2pnt(s.data,xMax), 0, DimSize(s.data,0)-1)
		endif
	endif
	
	if (isImage || isContour)
		if (rangeMode == 1)
			yMin = csrA_vals[3]
			yMax = csrB_vals[3]
		elseif (rangeMode == 2)
			if (isImage)
				GetAxis/W=$gName/Q $StringByKey("YAXIS",image)
			else
				GetAxis/W=$gName/Q $StringByKey("YAXIS",TraceInfo(gName,trace,0))
			endif
			yMin = !V_flag ? V_min : yMin
			yMax = !V_flag ? V_max : yMax
		endif
		if (!numtype(yMin) && !numtype(yMax))
			s.pntMinY = limit(ScaleToIndex(s.data,yMin,1), 0, DimSize(s.data,1)-1)
			s.pntMaxY = limit(ScaleToIndex(s.data,yMax,1), 0, DimSize(s.data,1)-1)
		endif
	endif
	
	if (s.pntMin > s.pntMax)
		temp = s.pntMin
		s.pntMin = s.pntMax
		s.pntMax = temp
	endif
	if (s.pntMinY > s.pntMaxY)
		temp = s.pntMinY
		s.pntMinY = s.pntMaxY
		s.pntMaxY = temp
	endif
	
	variable reverseX = 1
	if (WaveExists(s.xw))
		s.xBase = s.xw[0][xCol]
		reverseX = sign(s.xw[DimSize(s.xw,0)-1][xCol]-s.xw[0][xCol])
	else
		s.xBase  = DimOffset(s.data,0)
		reverseX = sign(DimDelta(s.data,0))
	endif
	
	if (isImage || isContour)
		WaveStats/Q/RMD=[s.pntMin,s.pntMax][s.pntMinY,s.pntMaxY] s.data
	else
		WaveStats/Q/RMD=[s.pntMin,s.pntMax][s.pntCol] s.data
	endif
		s.yBase = V_max > V_min ? V_min : V_max													// the baseline value
	if (rangeMode == 1 && !numtype(xMin) && !numtype(xMax))
		s.xBase = xMin*reverseX < xMax*reverseX ? xMin : xMax
	endif
	
	// ++++++++++++++++++++++++++++++++++++ do the fit +++++++++++++++++++++++++++++++++++++++++++
	string printStr = "", execStr = "", cwStr = ""
	if (StringMatch(fitType, "FuncFit"))
		FUNCREF dummyFunc_prepareCoef coefFunc=$(fitFunc+CoefGen_SUFFIX)						// prepare coef wave
		int error = coefFunc(s)
		if (error || !WaveExists(s.cw))
			Print "Guess for "+fitFunc+" failed."
			return -10
		endif
		
		printStr = "•Make/D/O/N="+num2str(DimSize(s.cw,0))+" W_coef={"
		for (i=0; i<DimSize(s.cw,0); i++)
			printStr += num2str(s.cw[i]) + ","
		endfor
		printStr = RemoveEnding(printStr,",")+"}"
		Print printStr																			// print coef creation command
		cwStr = GetWavesDataFolder(s.cw,2)+" "
	else
		K0 = s.holdMode == 1 ? 0 : s.yBase
		Printf "•K0 = %g;\r", K0
	endif
	
	string errBarInfo = StringByKey("ERRORBARS",TraceInfo(gName, trace,0))						// find error bar wave
	if (strlen(errBarInfo))
		sscanf StringByKey("wave",errBarInfo,"=",","), "(%s,", execStr
		wave/Z ebWave = $execStr
	endif
	
	string dataName = "", ebName = "", xwName = "", ywName = "", fitName = "fit_"+NameOfWave(s.data)
	if (WaveExists(ebWave) && doWeight)															// add x-wave and error info
		sPrintf ebName, "/W=%s[%d,%d]",GetWavesDataFolder(ebWave,2), s.pntMin, s.pntMax
	endif
	if (WaveExists(s.xw))
		sPrintf xwName, "/X=%s[%d,%d]%s",GetWavesDataFolder(s.xw,2), s.pntMin, s.pntMax, SelectString(isXTraceCol,"","["+Xcol_label+"]")
	endif
	if (isImage || isContour)																	// the data to fit
		sPrintf dataName, "%s[%d,%d][%d,%d]",GetWavesDataFolder(s.data,2), s.pntMin, s.pntMax, s.pntMinY, s.pntMaxY
	else
		sPrintf dataName, "%s[%d,%d]%s",GetWavesDataFolder(s.data,2), s.pntMin, s.pntMax, SelectString(isTraceOf2D,"","["+Ycol_label+"]")
	endif
	
	wave/Z fit = $fitName
	if (WaveExists(fit))																		// clear note to append derived values later
		Note/K $fitName
	endif
	
	string doTxtbox	= SelectString(s.doTextbox,"","/TBOX="+num2str(textboxOpt))					// construct the fit command
	string dofullR	= SelectString(s.plotFull,"","/X=1")
	string doHold	= SelectString(s.holdMode>0,"","/H=\""+s.holdStr+"\"")
	sPrintf execStr, "•%s/M=2%s%s%s %s %s%s /D%s%s\r", fitType, doTxtbox, dofullR, doHold, fitFunc, cwStr, dataName, xwName, ebName
	Print execStr
	Execute/Z execStr																			// do the fit
	
	wave/Z fit = $fitName
	if (V_flag)
		Print "Execution of the fit failed."
		if (WaveExists(fit))																	// invalidate any previous fit curve
			fit = NaN
		endif
		return -999
	endif

	if (StringMatch(fitType, "CurveFit"))
		wave s.cw = W_coef
	endif
	
	if (StringMatch(fitFunc, "Line"))															// the line function does not output M_Covar
		wave s.sw = W_sigma
	else
		wave s.sw = M_Covar
	endif
	
	// +++++++++++++++++++++++++++++++++++ post processing ++++++++++++++++++++++++++++++++++++++
	if (Exists(fitFunc+derivedVal_SUFFIX) == 6)
		FUNCREF dummyFunc_derivedVals outputFunc=$(fitFunc+derivedVal_SUFFIX)					// output additional parameters
		printStr = outputFunc(s)
		if (strlen(printStr))
			Print printStr
		endif
		if (WaveExists(fit))
			Note fit, printStr
		endif
	endif
	
	// +++++++++++++++++++++++++++++++++++++++ cleanup +++++++++++++++++++++++++++++++++++++++++++
	string remList = ""
	if (doCleanUp)																				// remove working variables and waves
		remList = "V_FitOptions;V_FitTol;V_tol;V_chisq;V_q;V_siga;V_sigb;V_Rab;V_Pr;V_r2;V_npnts;V_nterms;V_nheld;V_numNaNs;V_numINFs;"
		remList+= "V_FitError;V_FitQuitReason;V_FitIterStart;V_FitMaxIters;V_FitNumIters;"
		remList+= "V_startRow;V_startCol;V_startLayer;V_startChunk;V_endRow;V_endCol;V_endLayer;V_endChunk;"
		for(i=0; i<ItemsInList(remList); i++)
			Killvariables/Z $StringFromList(i,remList)
		endfor
		remList = "S_Info;"
		for(i=0; i<ItemsInList(remList); i++)
			Killstrings/Z $StringFromList(i,remList)
		endfor
		KillWaves/Z W_coef, W_sigma, M_Covar, W_ParamConfidenceInterval
	endif
	
	if (!WaveExists(fit) || isImage || isContour)
		return 0
	endif
	// +++++++++++++++++++++++++++++++++ trace related options +++++++++++++++++++++++++++++++++++
	string fitTrace = PossiblyQuoteName(NameOfWave(fit)), fitXTrace = ""
	int isPlotted  	= FindListItem(fitTrace, TraceNameList(gName,";",1)) != -1
	int wasPlotted 	= FindListItem(fitTrace, traceList) != -1
	
	if (isPlotted)
		fitXTrace = PossiblyQuoteName(StringByKey("XWAVE",TraceInfo(gName, fitTrace, 0)))
		if (strlen(fitXTrace))
			fitXTrace[0] = " vs "
		endif
	endif
	
	if (isTraceOf2D && doMultiOut)																// ver. 1.09 - create individual fit for each column
		sPrintf fitName, "fit_%.*d_%s", strlen(num2str(DimSize(s.data,1)-1)), s.pntCol, NameOfWave(s.data)	// the number of digits is decided by the order of magnitde of the size
		RemoveFromGraph/Z/W=$gName $fitTrace
		Duplicate/O fit, $fitName; KillWaves/Z fit
		fitTrace   = PossiblyQuoteName(fitName)
		isPlotted  = FindListItem(fitTrace, TraceNameList(gName,";",1)) != -1
		wasPlotted = FindListItem(fitTrace, traceList) != -1
	endif
	
	string fitTraceFlags = StringByKey("AXISFLAGS",TraceInfo(gName, fitTrace, 0))				// ver. 1.09 - put fit trace on the same axes as source trace
	string inpTraceFlags = StringByKey("AXISFLAGS",TraceInfo(gName, trace, 0))
	if (CmpStr(fitTraceFlags,inpTraceFlags) != 0 || !isPlotted)									// if the source is on a different axis combination, re-add fit trace to the same axes
		RemoveFromGraph/Z/W=$gName $fitTrace
		sPrintf execStr, "AppendToGraph/W=%s%s %s%s", gName, inpTraceFlags, fitTrace, fitXTrace
		Execute/Z/Q execStr
		isPlotted = 1
	endif
	
	if (isPlotted)
		if (!wasPlotted || (GetkeyState(0) & 2^0) != 0)											// is a new fit trace or ctrl / command was pressed
			sPrintf execStr, "ModifyGraph/W=%s lstyle(%s)=%d, lsize(%s)=%g", gName, fitTrace, setFitStyle, fitTrace, FitLineSize
			string colorMode = "", inputTrColor = ""
			Switch (setFitColor)	// 0 is default
				case 4:				// lower saturation
					colorMode = SelectString(strlen(colorMode),"lowerSaturation",colorMode)
				case 3:				// complementary color
					colorMode = SelectString(strlen(colorMode),"complementColor",colorMode)
				case 2:				// same as input trace
					inputTrColor = StringByKey("rgb(x)", TraceInfo(gName,trace,0), "=")
					inputTrColor = ReplaceString("(",RemoveEnding(inputTrColor,")"),"")			// remove outer brackets
					inputTrColor = AnalysisTools#RGB_modify(colorMode, inputTrColor)
					sPrintf execStr, "%s ,rgb(%s)=(%s)", execStr, fitTrace, inputTrColor
				break
				case 1:				// fixed color
					sPrintf execStr, "%s ,rgb(%s)=(%s)", execStr, fitTrace, FitColor
				break
			EndSwitch
			Execute/Z/Q execStr
		endif
		
		if (doFitAlign)																			// ver. 1.06 - added option for aligned fit traces
			variable xOff = 0, yOff = 0, xScale = 0, yScale = 0									// get scaling of source trace
			sscanf StringByKey("offset(x)", 	TraceInfo(gName,trace,0), "="), "{%f,%f}", xOff, yOff
			sscanf StringByKey("muloffset(x)",	TraceInfo(gName,trace,0), "="), "{%f,%f}", xScale, yScale
			ModifyGraph/Z/W=$gName offset($fitTrace)={xOff,yOff}, muloffset($fitTrace)={xScale,yScale}
		endif
	endif
	
	return 0
end

//################################################################################################

static Function getCol(string gName, string trace, int which, string &colStr)					// for traces of 2D data: extract column info
	string info
	if (which)
		wave/Z data = XWaveRefFromTrace(gName, trace)
		info = "XRANGE"
	else
		wave/Z data = TraceNameToWaveRef(gName, trace)
		info = "YRANGE"
	endif

	if (!WaveExists(data))
		return NaN
	endif
	if (WaveDims(data) < 2)
		return NaN
	endif
	
	variable colNum = NaN
	SplitString/E=("(?:(?:]\[))(.*)]$") StringByKey(info, TraceInfo(gName,trace,0)), colStr		// find the displayed column from TraceInfo
	if (strlen(colStr))
		if (!CmpStr(colStr[0],"%"))																// is tag
			colNum = FindDimLabel(data, 1, colStr[1,inf])
			colNum = colNum == -2 ? NaN : colNum
		else
			colNum = str2num(colStr)
		endif
	endif
	return colNum
end

//################################################################################################

static function/S RGB_modify(String which, String color)
	int R = str2num(StringFromList(0,color,",")); R = R < 0 ? 0 : R
	int G = str2num(StringFromList(1,color,",")); G = G < 0 ? 0 : G
	int B = str2num(StringFromList(2,color,",")); B = B < 0 ? 0 : B
	int A = str2num(StringFromList(3,color,",")); A = A <= 0 ? 65535 : A
	string returnColor = ""
	StrSwitch(which)
		case "complementColor":
			R = 65535-R;	G = 65535-G;	B = 65535-B;										// calculate complemetary colors
			if (min(R,G,B) >= 235*257)															// color is too bright to see
				R -= 15*257; G -= 15*257; B -= 15*257;											// make a bit darker
			endif
		break
		case "lowerSaturation":
			variable H,S,V
			[H,S,V] = AnalysisTools#RGB2HSV(R,G,B);		S *= 0.5								// extract and reduce saturation
			[R,G,B] = AnalysisTools#HSV2RGB(H,S,V)
		break
	EndSwitch
	sPrintf returnColor, "%d,%d,%d,%d", R, G, B, A
	return returnColor
end

//################################################################################################

static Constant kBitDepth = 65535																// 16 bit colors

static function [variable H, variable S, variable V] RGB2HSV(int R, int G, int B)				// assumes R,G,B values in the range [0..65535]
	variable Mx = max(R,G,B), Mn = min(R,G,B)
	H = acos((R-G/2-B/2)/sqrt(R^2+G^2+B^2-R*G-R*B-G*B))*180/pi
	H = G>=B ? 360-H : H; H = numtype(H)!=0 ? 0 : H
	S = Mx==0 ? 0 : (1-Mn/Mx)
	V = Mx/kBitDepth
	return [H,S,V]																				// outputs H = [0..360], S,V = [0..1]
end

static function [int R, int G, int B] HSV2RGB(variable H, variable S, variable V)				// assumes a H value in the range [0..360] and S,V values in the range [0..1]
	variable C = V*kBitDepth, M = C*(1-S), X = (C-M)*(1-abs(mod(H/60,2)-1))
	if     (H >=   0 && H < 60)
		R = C; G = M; B = X+M;
	elseif (H >=  60 && H < 120)
		R = X+M; G = M; B = C;
	elseif (H >= 120 && H < 180)
		R = M; G = X+M; B = C;
	elseif (H >= 180 && H < 240)
		R = M; G = C; B = X+M;
	elseif (H >= 240 && H < 300)
		R = X+M; G = C; B = M;
	elseif (H >= 300 && H <= 360)
		R = C; G = X+M; B = M;
	endif
	return [R,G,B]																				// outputs R,G,B = [0..65535]
end

//________________________________________________________________________________________________
//									Default Fit-Function Output
//________________________________________________________________________________________________

function/S Line_derivedVals(STRUCT SuperQuickFitStruct &s)
	NVAR V_rab
	variable xCross = -s.cw[0]/s.cw[1]
	variable xCross_err	= xCross * sqrt( (s.sw[0]/s.cw[0])^2 + (s.sw[1]/s.cw[1])^2 + 2*(V_rab*s.sw[0]*s.sw[1])/(s.cw[0]*s.cw[1]) )
	xCross_err = numtype(xCross_err) != 0 ? 0 : xCross_err
	string printStr = ""
	sPrintf printStr, "Derived values:\r\tX-intercept\t= %g ± %g", xCross, xCross_err
	if (s.doTextbox)
		AppendText "\t"+printStr
	endif
	return printStr
end

function/S Gauss_derivedVals(STRUCT SuperQuickFitStruct &s)
	variable pArea	= s.cw[1]*abs(s.cw[3])*sqrt(Pi)
	variable FWHM	= abs(s.cw[3])*2*sqrt(ln(2))
	variable pArea_err	= pArea * sqrt( (s.sw[3][3]/s.cw[3]^2) + (s.sw[1][1]/s.cw[1]^2) + 2*s.sw[1][3]/(s.cw[1]*s.cw[3]) )
	variable FWHM_err	= sqrt(s.sw[3][3])*2*sqrt(ln(2))
	string printStr = ""
	sPrintf printStr, "Derived values:\r\tArea\t= %g ± %g\r\tFWHM\t= %g ± %g", pArea, pArea_err, FWHM, FWHM_err
	if (s.doTextbox)
		AppendText "\t"+printStr
	endif
	return printStr
end

function/S Lor_derivedVals(STRUCT SuperQuickFitStruct &s)
	variable Height	= s.cw[1]/abs(s.cw[3])
	variable FWHM	= 2*sqrt(abs(s.cw[3]))
	variable pArea	= pi*s.cw[1]/sqrt(abs(s.cw[3]))
	variable Height_err	= Height * sqrt( (s.sw[3][3]/s.cw[3]^2) + (s.sw[1][1]/s.cw[1]^2) + 2*s.sw[1][3]/(s.cw[1]*s.cw[3]) )
	variable FWHM_err	= FWHM/s.cw[3] * sqrt(s.sw[3][3])
	variable pArea_err	= sqrt( s.sw[1][1] * pi^2/abs(s.cw[3]) + s.sw[3][3] * (pi^2*s.cw[1]^2)/(4*abs(s.cw[3])^3) - 2*abs(s.sw[1][3])*(pi*s.cw[1])/(2*abs(s.cw[3])^2) )
	string printStr = ""
	sPrintf printStr, "Derived values:\r\tHeight\t= %g ± %g\r\tArea\t= %g ± %g\r\tFWHM\t= %g ± %g", Height, Height_err, pArea, pArea_err, FWHM, FWHM_err
	if (s.doTextbox)
		AppendText "\t"+printStr
	endif
	return printStr
end

function/S Voigt_derivedVals(STRUCT SuperQuickFitStruct &s)
	variable FWHM_err = 0, Height_err = 0, C1 = 0.5346, C2 = 0.2166, k = 2*sqrt(ln(2)/pi)
	variable shape	= s.cw[4]*sqrt(ln(2)), lorW = s.cw[4]*s.cw[3]
	
	variable height	= k*s.cw[1]/abs(s.cw[3]) * exp(shape^2)*erfc(shape)
	variable FWHM = (C1*s.cw[4] + sqrt(C2*s.cw[4]^2 + 1))*abs(s.cw[3])

	Height_err +=   s.sw[1][1] * (  k/abs(s.cw[3])*exp(shape^2)*erfc(shape) )^2
	Height_err +=   s.sw[2][2] * ( (sqrt(pi)*shape*exp(shape^2)*erfc(shape) - 1)*k^2*s.cw[1]/abs(s.cw[3]) )^2
	Height_err +=   s.sw[3][3] * (  k*s.cw[1]/abs(s.cw[3])^2*exp(shape^2)*erfc(shape) )^2
	Height_err += 2*s.sw[1][3] *  k^2*s.cw[1]/abs(s.cw[3])^3*(exp(shape^2)*erfc(shape))^2
	Height_err -= 2*s.sw[1][2] *  (sqrt(pi)*shape*exp(shape^2)*erfc(shape) - 1)*k^3*s.cw[1]  /abs(s.cw[3])   *exp(shape^2)*erfc(shape)
	Height_err -= 2*s.sw[2][3] *  (sqrt(pi)*shape*exp(shape^2)*erfc(shape) - 1)*k^3*s.cw[1]^2/abs(s.cw[3])^2 *exp(shape^2)*erfc(shape)
	Height_err = sqrt(Height_err)
	
	FWHM_err += s.sw[3][3] * (  C1*s.cw[4] + sqrt(C2*s.cw[4]^2 + 1) )^2
	FWHM_err += s.sw[4][4] * ( (C1*s.cw[4] + C2*s.cw[4]/sqrt(C2*s.cw[4]^2 + 1))*abs(s.cw[3]) )^2
	FWHM_err += 2*abs(s.sw[3][4])*(C1*s.cw[4] + sqrt(C2*s.cw[4]^2 + 1))*(C1*s.cw[4] + C2*s.cw[4]/sqrt(C2*s.cw[4]^2 + 1))*abs(s.cw[3])
	FWHM_err = sqrt(FWHM_err)
	
	string printStr = ""
	sPrintf printStr, "Derived values:\r\tHeight\t= %g ± %g\r\tFWHM\t= %g ± %g", Height, Height_err, FWHM, FWHM_err
	if (s.doTextbox)
		AppendText "\t"+printStr
	endif
	return printStr
end

//________________________________________________________________________________________________
//									Custom Curve Fit Functions
//________________________________________________________________________________________________

function FermiEdge_Line(wave w, variable E) : FitFunc
	//CurveFitDialog/ Equation:
	//CurveFitDialog/ f(E) = y0 + (m*E+A) / (exp((E-Ef)/kT)+1)
	//CurveFitDialog/ 
	//CurveFitDialog/ End of Equation
	//CurveFitDialog/ Independent Variables 1
	//CurveFitDialog/ E
	//CurveFitDialog/ Coefficients 5
	//CurveFitDialog/ w[0] = y0
	//CurveFitDialog/ w[1] = Ef
	//CurveFitDialog/ w[2] = kT
	//CurveFitDialog/ w[3] = A
	//CurveFitDialog/ w[4] = m
	return w[0] + (w[4]*E+w[3]) / (exp((E-w[1])/w[2])+1)
end

function FermiEdge_Line_prepareCoef(STRUCT SuperQuickFitStruct &s)
	Duplicate/free/RMD=[][s.pntCol] s.data, tempDifWave											// support for 2D trace data
	Differentiate tempDifWave/D=tempDifWave
	variable smthFact = round(abs(s.pntMax-s.pntMin)/25)
	smthFact = limit(smthFact, 1, 2^15-1)
	Smooth smthFact, tempDifWave
	WaveStats/Q/R=[s.pntMin,s.pntMax] tempDifWave
	variable Ef
	if (WaveExists(s.xw))
		Ef = s.xw[V_minRowLoc]
	else
		Ef = V_minloc
	endif
	Make/D/O/N=2 W_coef = {(s.holdMode == 1 ? 0 : s.yBase),Ef,0.025,WaveMax(s.data),-1}
	wave s.cw = W_coef
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

function LineX(wave w, variable x) : FitFunc
	//CurveFitDialog/ Equation:
	//CurveFitDialog/ f(x) = m*(x-x0)
	//CurveFitDialog/ End of Equation
	//CurveFitDialog/ Independent Variables 1
	//CurveFitDialog/ x
	//CurveFitDialog/ Coefficients 2
	//CurveFitDialog/ w[0] = x0
	//CurveFitDialog/ w[1] = m
	return w[1]*(x-w[0])
end

function LineX_prepareCoef(STRUCT SuperQuickFitStruct &s)
	Duplicate/free/RMD=[][s.pntCol] s.data, tempDifWave											// support for 2D trace data
	Differentiate tempDifWave/D=tempDifWave
	WaveStats/Q/R=[s.pntMin,s.pntMax] tempDifWave
	variable slope = (WaveMax(s.data) - WaveMin(s.data)) > 0 ? V_max : V_min
	Make/D/O/N=2 W_coef = {(s.holdMode == 1 ? 0 : s.xBase),slope}
	wave s.cw = W_coef
	return 0
end

function/S LineX_derivedVals(STRUCT SuperQuickFitStruct &s)
	variable yCross = -s.cw[1]*s.cw[0]
	variable yCross_err	= yCross * sqrt( (s.sw[0][0]/s.cw[0]^2) + (s.sw[1][1]/s.cw[1]^2) + 2*s.sw[0][1]/(s.cw[0]*s.cw[1]) )
	string printStr = ""
	sPrintf printStr, "Derived values:\r\tY-intercept\t= %g ± %g", yCross, (numtype(yCross_err) != 0 ? 0 : yCross_err)
	if (s.doTextbox)
		AppendText "\t"+printStr
	endif
	return printStr
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

function PAD_Beta(wave w, variable x) : FitFunc
	//CurveFitDialog/ Equation:
	//CurveFitDialog/ f(theta) = scale*(1+beta*legendreA(2,0,cos((theta-theta_0)*(Pi/180))))
	//CurveFitDialog/ End of Equation
	//CurveFitDialog/ Independent Variables 1
	//CurveFitDialog/ theta
	//CurveFitDialog/ Coefficients 3
	//CurveFitDialog/ w[0] = scale
	//CurveFitDialog/ w[1] = beta
	//CurveFitDialog/ w[2] = theta_0
	return w[0]*(1+w[1]*legendreA(2,0,cos((x-w[2])*(Pi/180))))
end

function PAD_Beta_prepareCoef(STRUCT SuperQuickFitStruct &s)
	s.holdStr = "001"
	Make/D/O/N=3 W_coef = {WaveMax(s.data)/3,2,0}
	wave s.cw = W_coef
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

function DoubleGauss(wave w, variable x) : FitFunc
	//CurveFitDialog/ Equation:
	//CurveFitDialog/ f(x) = y0 + Area1 * Gauss(x,Loc1,Width1) + Area2 * Gauss(x,Loc2,Width2)
	//CurveFitDialog/ End of Equation
	//CurveFitDialog/ Independent Variables 1
	//CurveFitDialog/ x
	//CurveFitDialog/ Coefficients 7
	//CurveFitDialog/ w[0] = y0
	//CurveFitDialog/ w[1] = Area1
	//CurveFitDialog/ w[2] = Area2
	//CurveFitDialog/ w[3] = Loc1
	//CurveFitDialog/ w[4] = Loc2
	//CurveFitDialog/ w[5] = Width1
	//CurveFitDialog/ w[6] = Width2
	return w[0]+ w[1] * Gauss(x,w[3],w[5]) + w[2] * Gauss(x,w[4],w[6])
end

function DoubleGauss_prepareCoef(STRUCT SuperQuickFitStruct &s)
	Duplicate/free/RMD=[][s.pntCol] s.data, tempWaveCol											// support for 2D trace data
	variable minVal = s.holdMode == 1 ? 0 : s.yBase
	wave peakList = AnalysisTools#doFindPeaks(tempWaveCol,s.pntMin,s.pntMax)
	if (DimSize(peakList,0) < 2)
		return -1
	endif
	variable pLoc1 = peakList[0][0], pWidth1 = peakList[0][1], pHeight1 = peakList[0][2]
	variable pLoc2 = peakList[1][0], pWidth2 = peakList[1][1], pHeight2 = peakList[1][2]
	if (WaveExists(s.xw))
		pLoc1 = s.xw[pLoc1]
		pLoc2 = s.xw[pLoc2]
		pWidth1 *= abs((s.xw[numpnts(s.xw)-1]-s.xw[0])/(numpnts(s.xw)-1))
		pWidth2 *= abs((s.xw[numpnts(s.xw)-1]-s.xw[0])/(numpnts(s.xw)-1))
	else
		pLoc1 = pnt2x(s.data,pLoc1)
		pLoc2 = pnt2x(s.data,pLoc2)
		pWidth1 *= abs(DimDelta(s.data,0))
		pWidth2 *= abs(DimDelta(s.data,0))
	endif
	WaveStats/Q/R=[s.pntMin,s.pntMax] tempWaveCol
	pHeight1 = pHeight1 == 0 || numtype(pHeight1) != 0 ? V_max/20 : pHeight1
	pHeight2 = pHeight2 == 0 || numtype(pHeight2) != 0 ? V_max/20 : pHeight2
	Make/D/O/N=7 W_coef = {minVal,pHeight1,pHeight2,pLoc1,pLoc2,pWidth1,pWidth2}
	wave s.cw = W_coef
	return 0
end

function/S DoubleGauss_derivedVals(STRUCT SuperQuickFitStruct &s)
	variable FWHM1		= abs(s.cw[5])*2*sqrt(2*ln(2)),			FWHM2		= abs(s.cw[6])*2*sqrt(2*ln(2))
	variable FWHM1_err	= sqrt(s.sw[5][5])*2*sqrt(2*ln(2)),		FWHM2_err	= sqrt(s.sw[6][6])*2*sqrt(2*ln(2))
	variable Amp1		= s.cw[1]/(abs(s.cw[5])*sqrt(2*Pi)),	Amp2		= s.cw[2]/(abs(s.cw[6])*sqrt(2*Pi))
	variable Amp1_err	= Amp1 * sqrt( (s.sw[5][5]/s.cw[5]^2) + (s.sw[1][1]/s.cw[1]^2) + 2*s.sw[1][5]/(s.cw[1]*s.cw[5]) )
	variable Amp2_err	= Amp2 * sqrt( (s.sw[6][6]/s.cw[6]^2) + (s.sw[1][1]/s.cw[1]^2) + 2*s.sw[1][6]/(s.cw[1]*s.cw[6]) )
	string printStr = ""
	sPrintf printStr, "Derived values:\r\tAmp1\t= %g ± %g\r\tAmp2\t= %g ± %g\r\tFWHM1\t= %g ± %g\r\tFWHM2\t= %g ± %g", Amp1, Amp1_err, Amp2, Amp2_err, FWHM1, FWHM1_err, FWHM2, FWHM2_err
	if (s.doTextbox)
		AppendText "\t"+printStr
	endif
	return printStr
end

static function/wave doFindPeaks(wave w, int p0, int p1)										// finds a list of possible peaks between points p0 and p1
	Variable resampleFactor = p1-p0 < 150 ? max(2,ceil(150/(p1-p0))) : 1
	Duplicate/FREE w, waveResampled
	Resample/UP=(resampleFactor)/WINF=None waveResampled
	p0 *= resampleFactor
	p1 *= resampleFactor
	SetScale/P x,0,1,waveResampled
	
	int smFact = limit((p1-p0)/50,2,8)															// smoothing factor scales with size
	Duplicate/Free/R=[p0,p1] waveResampled,waveSmoothed,waveDoubleDif; 	Smooth/E=2/B=3 smFact,waveSmoothed,waveDoubleDif
	Differentiate waveDoubleDif;					 					Smooth/E=2/B=3 2*smFact, waveDoubleDif
	Differentiate waveDoubleDif;										Smooth/E=2/B=3 2*smFact, waveDoubleDif
	
	Make/D/N=(0,4)/free peakResults
	int pStart=0, wn=DimSize(waveDoubleDif,0)-1
	do
		findpeak/B=1/N/P/R=[pStart,]/Q waveDoubleDif; variable pC=V_PeakLoc, mVal=V_PeakVal
		if (V_flag)
			break
		endif
		findpeak/B=1/P/R=[pC,0]/Q waveDoubleDif; variable pL=V_flag ? 0  : V_PeakLoc, lVal=V_flag ? waveDoubleDif[0]  : V_PeakVal
		findpeak/B=1/P/R=[pC, ]/Q waveDoubleDif; variable pR=V_flag ? wn : V_PeakLoc, rVal=V_flag ? waveDoubleDif[pR] : V_PeakVal
		variable height = 1.3*(waveSmoothed[pC]-min(waveSmoothed[pR],waveSmoothed[pL]))
		peakResults[DimSize(peakResults,0)][]={{pC+p0}, {abs(pR-pL)/sqrt(6)}, {height}, {min(abs(mVal-lVal),abs(mVal-rVal))}}
		if (V_flag)
			break
		endif
		pStart=pR
	while(1)
	if (DimSize(peakResults,0) > 0)																// sort from strongest to weakest peak
		Make/D/N=(DimSize(peakResults,0))/FREE heightSortWave = peakResults[p][3]
		MakeIndex/R heightSortWave, heightSortWave
		Duplicate/Free peakResults, peakResultsCP
		peakResults = peakResultsCP[heightSortWave[p]][q]
	endif
	peakResults[][0,1] /= resampleFactor
	return peakResults
end

// ####################################### Helper functions ######################################

static function getShapeFromData(STRUCT SuperQuickFitStruct &s,variable &loc, variable &amp, variable &base, variable &lwidth, variable &rwidth)	// finds basic parameters like hieght, width and location from input data
	Duplicate/free/RMD=[][s.pntCol] s.data, tempData											// support for 2D trace data
	WaveStats/Q/R=[s.pntMin,s.pntMax] tempData
	amp  = abs(V_max) > abs(V_min) ? V_max-V_min : V_min-V_max
	base = s.holdMode == 1 ? 0 : s.yBase
	if (WaveExists(s.xw))
		loc = amp > 0 ? s.xw[V_maxRowloc] : s.xw[V_minRowloc]
	else
		loc = amp > 0 ? V_maxloc : V_minloc
	endif
	lwidth = NaN; rwidth = NaN
	// +++++++++++++++++++++++++++++++++ width and asymmetry +++++++++++++++++++++++++++++++++++++
	Make/FREE/D/N=2 Levels
	variable reverseX, avg = ceil((s.pntMax-s.pntMin)*0.05)
	if(WaveExists(s.xw))
		reverseX = sign(s.xw[numpnts(s.xw)-1]-s.xw[0])
		FindLevels/Q/R=[s.pntMin,s.pntMax]/D=Levels/B=(limit(avg,1,20))/P tempData, amp/2+base	// find the half-maximum positions
		if (V_LevelsFound == 2)
			if (Levels[0] > 0 && Levels[1] > 0 && Levels[0] < numpnts(s.xw) && Levels[1] < numpnts(s.xw))
				lwidth = (loc - s.xw[Levels[0]])*reverseX
				rwidth = (s.xw[Levels[1]] - loc)*reverseX
			endif
		endif
	else
		reverseX = sign(DimDelta(s.data,0))
		FindLevels/Q/R=[s.pntMin,s.pntMax]/D=Levels/B=(limit(avg,1,20))   tempData, amp/2+base
		if (V_LevelsFound == 2)
			lwidth = (loc - Levels[0])*reverseX
			rwidth = (Levels[1] - loc)*reverseX
		endif
	endif
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function getPeakParameters(string PeakFunc, wave cw, variable &amp, variable &loc, variable &fwhm)	// outputs peak height, location and FWHM from a set of initial guesses
	variable isPos = amp > 0, start_w = fwhm
	Optimize/Q/A=(isPos)/L=(loc-5*start_w)/H=(loc+5*start_w) $PeakFunc,cw						// find peak maximum numerically
	amp  = isPos ? V_max : V_min
	loc  = isPos ? V_maxLoc : V_minLoc
	fwhm = AnalysisTools#getFWHM(PeakFunc,cw, amp, loc, start_w)
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function getFWHM(string PeakFunc, wave cw, variable amp, variable loc, variable start_w)	// finds the approximate FWHM numerically
	variable highRange, lowRange, highHW, LowHW
	highRange = amp > 0 ? loc : loc+5*start_w
	lowRange  = amp > 0 ? loc+5*start_w : loc
	FindRoots/Q/H=(highRange)/L=(lowRange)/T=1e-15/Z=(amp/2) $PeakFunc,cw
	highHW = V_flag ? NaN : V_Root
	highRange = amp > 0 ? loc : loc-5*start_w
	lowRange  = amp > 0 ? loc-5*start_w : loc
	FindRoots/Q/H=(highRange)/L=(lowRange)/T=1e-15/Z=(amp/2) $PeakFunc,cw
	LowHW = V_flag ? NaN : V_Root
	return abs(highHW - LowHW)
end

// ####################################### EMG Peak function #####################################

function ExpModGauss(wave cw, wave yw, wave xw)													// the fit function
	Make/FREE/D/N=4 peak_cw = cw[p]
	variable dummy = MPFXEMGPeak(peak_cw, yw, xw)
	yw += cw[4]
end
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function ExpModGauss_prepareCoef(STRUCT SuperQuickFitStruct &s)									// the coef generator
	Make/D/O/N=5 W_coef	= NaN		// [0]: x0, [1]: width, [2]: amplitude, [3]: asymmetry, [4]: y0
	variable, loc, amp, base, lwidth, rwidth
	AnalysisTools#getShapeFromData(s, loc, amp, base, lwidth, rwidth)							// gets location, amplitude, baseline and width from the data
	
	W_coef[0] = loc
	W_coef[2] = amp
	W_coef[4] = base
	if (numtype(lwidth+rwidth) == 0)															// half-maximum positions found
		W_coef[0] += (lwidth - rwidth)/2														// move the peak over a bit
		if (lwidth > rwidth)																	// left tail is longer
			W_coef[1] = rwidth
			W_coef[3] = -3*abs(lwidth - rwidth)													// updated ExpTau calculation to work well with asymmetric peak edit
			W_coef[3] = min(W_coef[3],-0.2*W_coef[1])											// ExpModGauss needs some minimal asymmetry
		else
			W_coef[1] = lwidth
			W_coef[3] = 3*abs(lwidth - rwidth)
			W_coef[3] = max(W_coef[3],0.2*W_coef[1])
		endif
		variable widthRatio = abs(W_coef[3])/W_coef[1]
		W_coef[2] = W_coef[2]/(1-(widthRatio/(widthRatio+1.2))^2)								// totally empirically determined
	else
		W_coef[1] = 1																			// may not be good guesses
		W_coef[3] = 0.2
	endif
	
	wave s.cw = W_coef
	s.holdStr = "00001"																			// may hold baseline
	return 0
end
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function/S ExpModGauss_derivedVals(STRUCT SuperQuickFitStruct &s)								// calculate derived parameters
	variable amp = s.cw[2], loc = s.cw[0], fwhm = abs(s.cw[1])									// must provide (rough) guesses
	AnalysisTools#getPeakParameters("AnalysisTools#EMGFunc", s.cw, amp, loc, fwhm)
	
	variable pArea = s.cw[1]*s.cw[2]*sqrt(2*pi)													// the peak area can be calculated analytically
	variable pArea_err = sqrt(s.sw[1][1]*(s.cw[2]*sqrt(2*pi))^2 + s.sw[2][2]*(s.cw[1]*sqrt(2*pi))^2 + 2*abs(s.sw[1][2])*(s.cw[1]*s.cw[2]*2*pi))
	
	string printStr = ""
	sPrintf printStr, "Area\t= %.5g ± %.5g\r\tHeight\t= %.7g\r\tLocation\t= %.7g\r\tFWHM\t= %.7g", pArea, pArea_err, amp, loc, fwhm
	if (s.doTextbox)
		AppendText "\t"+printStr																// also add derived parameters into text-box
	endif
	Print "K0 = x0, K1 = width, K2 = amplitude, K3 = asymmetry, K4 = y0"						// print explanation
	return "Derived values:\r\t"+printStr
end
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static function EMGFunc(wave w, variable x)														// function for finding peak values
	Make/D/FREE/N=1 ytemp, xtemp=x
	variable dummy = MPFXEMGPeak(w, ytemp, xtemp)
	return ytemp[0]
end

// ####################################### ExpConvExp function #####################################

function ExpConvExp(wave cw, wave yw, wave xw)													// the fit function
	Make/FREE/D/N=4 peak_cw = cw[p]
	variable dummy = MPFXExpConvExpPeak(peak_cw, yw, xw)
	yw += cw[4]
end
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function ExpConvExp_prepareCoef(STRUCT SuperQuickFitStruct &s)									// the coef generator
	Make/D/O/N=5 W_coef	= NaN	// [0]: x0, [1]: height, [2]: 1st inverse decay constant, [3]: 2nd inverse decay constant, [4]: y0
	variable, loc, amp, base, lwidth, rwidth
	AnalysisTools#getShapeFromData(s, loc, amp, base, lwidth, rwidth)
	
	W_coef[0] = loc
	W_coef[1] = amp
	W_coef[4] = base
	if (numtype(lwidth+rwidth) == 0)															// half-maximum positions found
		W_coef[2] = 1/lwidth
		W_coef[3] = 1/rwidth
	else
		W_coef[2] = 1/(0.04*(s.pntMax-s.pntMin)*abs(DimDelta(s.data,0)))						// may not be good guesses
		W_coef[3] = 1/(0.06*(s.pntMax-s.pntMin)*abs(DimDelta(s.data,0)))
	endif
	if (W_coef[2] >= W_coef[3])																	// coefficients should not be exactly the same
		W_coef[3] = 1.2*W_coef[2]																// make sure that decay coefficients are in the right order
	endif
	W_coef[1] *= 1.6																			// make the peak a bit higher
	W_coef[0] -= (ln(W_coef[2]/W_coef[3])/(W_coef[2]-W_coef[3]))								// shift the peak over
	
	wave s.cw = W_coef
	s.holdStr = "00001"																			// may hold baseline
	return 0
end
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function/S ExpConvExp_derivedVals(STRUCT SuperQuickFitStruct &s)								// calculate derived parameters
	variable amp, loc, fwhm
	AnalysisTools#getExpConvExpVals(s.cw, amp, loc, fwhm)

	variable pArea = s.cw[1]/s.cw[3]															// the peak area can be calculated analytically
	variable pArea_err = sqrt(s.sw[1][1]/(s.cw[3])^2 + s.sw[3][3]*(s.cw[1]/s.cw[3]^2)^2 + 2*abs(s.sw[1][3])*(s.cw[1]/s.cw[3]^3))

	variable K3 = s.cw[2], K3_err = sqrt(s.sw[2][2])
	variable K2_mod = s.cw[3]-s.cw[2]															// add alternative coefficients
	variable K2_err = sqrt(s.sw[2][2] + s.sw[3][3] + 2*abs(s.sw[2][3]))
	variable K1_mod = (s.cw[2]*s.cw[1])/K2_mod
	
	variable d1 = s.cw[2]/K2_mod, d2 = (s.cw[3]*s.cw[1])/K2_mod^2, d3 = (s.cw[2]*s.cw[1])/K2_mod^2		// derivatives
	variable K1_err = s.sw[1][1]*(d1)^2 + s.sw[2][2]*(d2)^2 + s.sw[3][3]*(d3)^2
	K1_err += 2*abs(s.sw[1][3])*d1*d3 + 2*abs(s.sw[1][2])*d1*d2 + 2*abs(s.sw[2][3])*d2*d3				// side terms
	K1_err = sqrt(K1_err)

	string baseStr = "", advStr = ""
	sPrintf baseStr, "Area\t= %.5g ± %.5g\r\tHeight\t= %.7g\r\tLocation\t= %.7g\r\tFWHM\t= %.7g", pArea, pArea_err, amp, loc, fwhm
	sPrintf advStr,  "Virt. amplitude\t= %.5g ± %.5g\r\tInv. of rise\t= %.5g ± %.5g\r\tInv. of decay\t= %.5g ± %.5g", K1_mod, K1_err, K2_mod, K2_err, K3, K3_err
	if (s.doTextbox)
		AppendText "\t"+baseStr + "\r\t-----\r\t" + advStr										// also add derived parameters into text-box
	endif
	Print "K0 = location, K1 = amplitude,  K2 = 1st inv. decay const., K3 = 2nd inv. decay const., K4 = baseline"	// print explanation
	return "Derived values:\r\t"+baseStr + "\r\t-----\r\t" + advStr	
end
// ++++++++++++++++++++++++++++++ recreated from Multipeak Fit +++++++++++++++++++++++++++++++++++
static function getExpConvExpVals(wave cw, variable &amp, variable &loc, variable &fwhm)
	if (cw[2] == cw[3])
		amp = cw[1]/e
		loc = 1/cw[3] + cw[0]
	else
		amp = (cw[1]*cw[2] * ((cw[2]/cw[3])^(-(cw[2]/(cw[2]-cw[3]))) - (cw[2]/cw[3])^(-(cw[3]/(cw[2] - cw[3]))))) / (-cw[2]+cw[3])
		loc = ln(cw[2]/cw[3])/(cw[2]-cw[3]) + cw[0]
	endif
	fwhm  = AnalysisTools#getFWHM("AnalysisTools#ExpConvExpFunc",cw, amp, loc, abs(1/cw[2]+1/cw[3]))
	return 0
end
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static function ExpConvExpFunc(wave w, variable x)												// function for finding peak values
	Make/D/FREE/N=1 ytemp, xtemp=x
	variable dummy = MPFXExpConvExpPeak(w, ytemp, xtemp)
	return ytemp[0]
end